/**************************************************************************
 *
 *-------------------------------------------------------------------------
 * Project : Short Read Assembler 
 *
 *-------------------------------------------------------------------------
 #######################################################################
 # This software has been created by Genome Research Limited (GRL).    #
 # GRL hereby grants permission to use, copy, modify and distribute    #
 # this software and its documentation for non-commercial purposes     #
 # without fee at the user's own risk on the basis set out below.      #
 # GRL neither undertakes nor accepts any duty whether contractual or  #
 # otherwise in connection with the software, its use or the use of    #
 # any derivative, and makes no representations or warranties, express #
 # or implied, concerning the software, its suitability, fitness for   #
 # a particular purpose or non-infringement.                           #
 # In no event shall the authors of the software or GRL be responsible #
 # or liable for any loss or damage whatsoever arising in any way      #
 # directly or indirectly out of the use of this software or its       #
 # derivatives, even if advised of the possibility of such damage.     #
 # Our software can be freely distributed under the conditions set out #
 # above, and must contain this copyright notice.                      #
 #######################################################################
 *
 *  Author : Yong Gu (ygu@sanger.ac.uk) 
 *
 *  Copyright (C) 2009 by Genome Research Limited, All rights reserved.
 *
 **************************************************************************/

#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <dirent.h>
#include <values.h>
#include <string.h>
#include <math.h>
#include <unistd.h>

#include "utils.h"
#include "fpathM.h"
#include "kmerP.h"
#include "kmerW.h"
#include "seqE.h"

void inputReadFastQ(char *fname, SeqDB *db)
{
	long length, remain, blockSize, nBlock, i, j;
	char *FQdata;
	char *ptr, *ptr2;
	FILE *fp;
	int numRead, fnLen;
	char *fname1;
	int readLen, readLen2;
	long FQsize;
	Fastq *read;
	long nBase;
	
	fp = openfile(fname,"r");
	fseek(fp, 0, SEEK_END);
	length = ftell(fp);
	FQsize = length;
	rewind(fp);

	blockSize = 1024;

	FQdata = (char *) malloc(sizeof(char)*length);
	nBlock = length/blockSize;
	remain = length - nBlock*blockSize;

	if((j=fread(FQdata, blockSize, nBlock, fp)) != nBlock){
		fprintf(stderr,"file reading %s error: block reading\n",fname);
		exit(1);
	}

	if(remain != 0){
		if((j=fread(FQdata+blockSize*nBlock, sizeof(char), remain, fp)) != remain){
			fprintf(stderr,"file reading %s error: remain reading\n",fname);
			exit(1);
		}
	}

        numRead = 0;
        ptr = FQdata;
        for(i=0;i<length;++i){
                if(*ptr++ == '\n')++numRead;
        }
        numRead /= 4;
        read = (Fastq *) malloc(sizeof(Fastq)* (numRead));

	ptr = FQdata;
	numRead = 0;
	nBase = 0;
	while(ptr-FQdata < length){
		if(*ptr++ != '@'){
			fprintf(stderr,"FQread error \n");
			exit(1);
		}
		fname1 = (char *)ptr;
		fnLen = 0;
		while(*ptr++ != '\n'){
			++fnLen;
		}
		ptr[-1] = '\0';
		read[numRead].name = strtok(fname1," ");
		read[numRead].base = (char *)ptr;
		ptr2 = ptr;
		readLen=0;
		while(*ptr != '+'){
			if(*ptr == '\n') {
				ptr++;
				continue;
			}
			*ptr2++ = *ptr++;
			readLen++;
		}
		*ptr2 = '\0';
		read[numRead].length = readLen;

		while(*ptr != '\n')++ptr;
		if(*ptr != '\n'){
			fprintf(stderr,"FQread error \n");
			exit(1);
		}
		*ptr++ = '\0';
		read[numRead].qual = (char *)ptr;
		readLen2 = 0;
		ptr2 = ptr;
                while( *ptr != '@' || readLen2 < readLen ){
			if(ptr == FQdata+length)break;
                        if(*ptr == '\n') {
                                ptr++;
                                continue;
                        }
                        *ptr2++ = *ptr++ - 041; 
			readLen2++;
                }
		*ptr2 = '\0';
		if(readLen2 != readLen){
	                fprintf(stderr,"ERROR: read %s contains different number of bases and quals\n", fname1);
			exit(1);
		}
		++numRead;
		nBase += readLen;
	}
	fclose(fp);
	db->nRead = numRead;
	db->read = read;
	db->FQdata = FQdata;
	db->nBase = nBase;
}


int seqExtending(char *fname, ExtCntlD *extCntlP, char *ofname)
{
	SeqDB seqD, *seqP; seqP = &seqD;
	FullKmerL *fullKmerP, *fullKmerD; 
	FullKmerL **fKmerSrt; 
	UniqKmerL *uniqKmerP;
	KmerDB kmerD, *kmerP; kmerP = &kmerD;
	long nFKmer;
	int i;
	int kmerL = extCntlP->kmerL;
	ReptKmerId reptKmerIdD, *reptKmerIdP; reptKmerIdP = &reptKmerIdD;

	initialising(seqP);
	inputReadFastQ(fname, seqP);
	kmerP->kmerL = kmerL;

	nFKmer = seqP->nBase - (kmerL-1)*seqP->nRead;
	fullKmerD = (FullKmerL *) malloc(sizeof(FullKmerL)*(nFKmer+2));
	fullKmerD[0].readId = -1;
	fullKmerP = fullKmerD+1;
	fKmerSrt = (FullKmerL **) malloc(sizeof(FullKmerL *)*nFKmer);
	kmerP->fKmer = fullKmerP;
	kmerP->fKmerSrt = fKmerSrt;
	kmerP->nFKmer = 0;

	read2kmerL(seqP, kmerP);
	fullKmerP[kmerP->nFKmer].readId = -1;

	ReadInf *readInfP = malloc(sizeof(ReadInf)*seqP->nRead);
	for(i=0;i<seqP->nRead;++i){
		readInfP[i].post = 0;
		readInfP[i].postR = -2;
		readInfP[i].kmerI = -1;
		readInfP[i].ctgId = -1;
		readInfP[i].ctgOffset = 0;
		if((i%2) == 0) readInfP[i].pairedId = i+1;
		else readInfP[i].pairedId = i-1;
	}

	sortKmerL(kmerP, readInfP, seqP->nRead);

	uniqKmerP = kmerP->uKmer;

	kmerP->lnkKmerL = (int *) malloc(sizeof(int)*2*kmerP->nFKmer); 
	lnkKmerLst(kmerP);
	shtPathRmv(kmerP, seqP->read, extCntlP->qThresh);
	shtPathRmv(kmerP, seqP->read, extCntlP->qThresh);

	extCntlP->cnsSeq = (char *) malloc(extCntlP->maxExtLen+2*kmerP->kmerL);

	extCntlP->numSetCover = 1.5 * (kmerP->totDepth/kmerP->totUsedKmer);
	extCntlP->nContig = 0;

	extCntlP->readList = (int *)  malloc(sizeof(int)*seqP->nRead);

	extCntlP->fpOut = openfile(ofname,"w");
	seqExtM(kmerP, seqP, readInfP, reptKmerIdP, extCntlP);
	seqExtR(kmerP, seqP, readInfP, reptKmerIdP, extCntlP);

	fclose(extCntlP->fpOut);
	return(1);
}

void usage(char *pname)
{
	fprintf(stderr, "Usage: %s [-f] [-d depth] [-k kmerLength] [-l extSeqLen] [-n numExtRead] [-i insertSize] -s readsFastqFile -o outputFile\n", pname);
        fprintf(stderr, "Options: -f :activated fuzzy check \n");
	fprintf(stderr, "         -d integer :depth setting (1500) \n");
        fprintf(stderr, "         -k integer :kmer length (27) \n");
        fprintf(stderr, "         -l integer :max length of sequence extension (3000) \n");
        fprintf(stderr, "         -n integer :number of sequences required for extension (1000) \n");
        fprintf(stderr, "         -i integer :insert size for read pairs (160) \n");
        fprintf(stderr, "         -s filename :input reads file (fastq) \n");
        fprintf(stderr, "         -o filename :output seqences file (fastq) \n");
}

int main(int argc, char **argv)
{
	char *readFile, *outFile;
	int opt;
	ExtCntlD extCntlD, *extCntlP; extCntlP=&extCntlD;

	if(argc<4){
		usage(argv[0]);
		return(0);
	}

	extCntlP->misMatch=5;	  extCntlP->nExtRead=0;	    extCntlP->useAllRead=0;
	extCntlP->fuzzyCheck=0;   extCntlP->kmerL=27;       extCntlP->depth=1500; 
	extCntlP->maxExtLen=3000; extCntlP->insertSize=160; extCntlP->pairCheck=0;
	extCntlP->qThresh=25;     extCntlP->sglHitThr=3;    
	// extCntlP->maxExtRead=50000;
	readFile=outFile=NULL;
	while ((opt = getopt(argc, argv, "fd:i:k:l:o:s:n:")) >= 0) {
		switch (opt) {
			case 'f': 
				extCntlP->fuzzyCheck = 1; 
				break;
			case 's':
				readFile = optarg;
				break;
			case 'o':
				outFile = optarg;
				break;
			case 'k':
				extCntlP->kmerL = atoi(optarg);
				break;
			case 'd':
				extCntlP->depth = atoi(optarg);
				break;
			case 'l':
				extCntlP->maxExtLen = atoi(optarg);
				break;
			case 'i':
				extCntlP->insertSize = atoi(optarg);
				if(extCntlP->insertSize > 0) extCntlP->pairCheck = 1;
				break;
			case 'n':
				extCntlP->nExtRead = atoi(optarg);
				break;
		}
	}
	if(readFile==NULL||outFile==NULL){
		usage(argv[0]);
		return(0);
	}

	if(seqExtending(readFile, extCntlP, outFile) == 1){
		fprintf(stderr, "%s has finished!\n", argv[0]);
	}
	else {
		fprintf(stderr, "There is an error in %s\n", argv[0]);
	}
	return(1);
}

