/**************************************************************************
 *
 *-------------------------------------------------------------------------
 * Project : Short Read Assembler 
 *
 *-------------------------------------------------------------------------
 #######################################################################
 # This software has been created by Genome Research Limited (GRL).    #
 # GRL hereby grants permission to use, copy, modify and distribute    #
 # this software and its documentation for non-commercial purposes     #
 # without fee at the user's own risk on the basis set out below.      #
 # GRL neither undertakes nor accepts any duty whether contractual or  #
 # otherwise in connection with the software, its use or the use of    #
 # any derivative, and makes no representations or warranties, express #
 # or implied, concerning the software, its suitability, fitness for   #
 # a particular purpose or non-infringement.                           #
 # In no event shall the authors of the software or GRL be responsible #
 # or liable for any loss or damage whatsoever arising in any way      #
 # directly or indirectly out of the use of this software or its       #
 # derivatives, even if advised of the possibility of such damage.     #
 # Our software can be freely distributed under the conditions set out #
 # above, and must contain this copyright notice.                      #
 #######################################################################
 *
 *  Author : Yong Gu (ygu@sanger.ac.uk) 
 *
 *  Copyright (C) 2009 by Genome Research Limited, All rights reserved.
 *
 **************************************************************************/

#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <dirent.h>
#include <values.h>
#include <string.h>
#include <math.h>
#include <unistd.h>

#include "utils.h"
#include "fpathM.h"
#include "kmerP.h"
#include "kmerW.h"

void kmer2base(long kmerC, char *cnsSeq, int kmerL, int rc, char *baseIndex)
{
	long k;
	int baseC;

	k = kmerC;

	if(rc==1){
		baseC = (int) (k>>(2*(kmerL-1)))&3L;
		*cnsSeq = baseIndex[baseC];
	}
	else {
		baseC = (int) k&3L;
		*cnsSeq = baseIndex[baseC+6];
	}
}

void kmer2seq(long kmerC, char *cnsSeq, int kmerL, int rc, char *baseIndex)
{
	int i; 
	long k;
	int baseC;

	k = kmerC;

	if(rc==1){
		for(i=0;i<kmerL;++i){
			baseC = (int) k&3L;
			cnsSeq[kmerL-i-1] = baseIndex[baseC];
			k = k>>2;
		}
	}
	else {
		for(i=0;i<kmerL;++i){
			baseC = (int) k&3L;
			cnsSeq[i] = baseIndex[baseC+6];
			k = k>>2;
		}
	}
}

int maskReptKmer(FullKmerL **fKmerSrt, int st, int end)
{
	int i, j;
	int masked, nMasked;

	nMasked = 0;
	i=st;
	while(i<end-1){
		masked = 0;
		j = i+1;
		while(j<end){
			if(fKmerSrt[i]->readId == fKmerSrt[j]->readId && fKmerSrt[i]->kmerC == fKmerSrt[j]->kmerC){
				fKmerSrt[j]->kmerC = -100;
				masked = 1;
				++nMasked;
				++j;
			}
			else break;
		}
		if(masked){
			fKmerSrt[i]->kmerC = -100;
			++nMasked;
		}
		i=j;
	}
	return(nMasked);
}

int kmerCPosCmp(const void *v1, const void *v2)
{
	long d;
	
	d = (*(FullKmerL **)v1)->kmerC - (*(FullKmerL **)v2)->kmerC;
	if(d > 0) return(1);
	if(d < 0) return(-1);
	return((*(FullKmerL **)v2)->readPos - (*(FullKmerL **)v1)->readPos);
}

typedef struct _hsList {
        int head;
        int size;
	int i;
} HSList;


void hSort(int wLen, int nShift, int nFKmer, FullKmerL **fKmerSrt, FullKmerL *fullKmerP)
{
	int n, i, cHead;
	HSList *hsList;
	int *highI;
	int nMasked=0;

	n = 1 << (wLen - nShift);

	hsList = (HSList *) malloc(n*sizeof(HSList));
	highI = (int *) malloc(nFKmer*sizeof(int));
	for(i=0;i<n;++i) hsList[i].size = 0;
	int *hPtr = highI;
	for(i=0;i<nFKmer;++i){
		*hPtr = fullKmerP[i].kmerC>>nShift;
		hsList[*hPtr++].size++;
	}
	cHead = 0;
	HSList *ptr = hsList;
	for(i=0;i<n;++i){
		ptr->i = cHead;
		ptr->head = cHead;
		cHead+=ptr++->size;
	}
	hPtr = highI;
	for(i=0;i<nFKmer;++i){
		fKmerSrt[hsList[*hPtr++].i++] = &fullKmerP[i];
	}
	for(i=0;i<n;++i){
		if(hsList[i].size == 0) continue;
		nMasked += maskReptKmer(fKmerSrt,hsList[i].head,hsList[i].head+hsList[i].size-1);
		qsort(&fKmerSrt[hsList[i].head], hsList[i].size, sizeof(FullKmerL *), kmerCPosCmp);
	}
	free(hsList);
	free(highI);
}

void sortKmerL(KmerDB *kmerP, ReadInf *readInfP, int nRead)
{
	int i, j, k;
	int l = 0; long kmerC; int ntmp;
	FullKmerL **fKmerSrt = kmerP->fKmerSrt;
	FullKmerL *fullKmerP = kmerP->fKmer;
	UniqKmerL *uniqKmerP;
	long totDepth=0;
	int totUsedKmer=0;
	int nShift, nSort;

	int *kmerDepth = (int *) malloc(sizeof(int)*kmerP->nFKmer);
	int *kmerDepthP;

	nSort = nRead;
	for(i=0;nSort>10;++i) nSort>>=1;
	nSort = 1L<<i;
	nShift = 2*kmerP->kmerL-i;
	hSort(2*kmerP->kmerL, nShift, kmerP->nFKmer, fKmerSrt, fullKmerP);

	i = 0;
	l = 0;
	kmerDepthP = kmerDepth;
	while(i<kmerP->nFKmer){
		kmerC = fKmerSrt[i]->kmerC;
		fKmerSrt[i]->uKmerId = l;
		for(j=i+1;j<kmerP->nFKmer; ++j){
			if(kmerC != fKmerSrt[j]->kmerC) break;
			fKmerSrt[j]->uKmerId = l;
		}
		*kmerDepthP++ = j - i;
		++l;
		i = j;
	}
	kmerP->nUKmer = l;
	
	kmerP->uKmer = (UniqKmerL *) malloc(sizeof(UniqKmerL)*kmerP->nUKmer); 
        uniqKmerP=kmerP->uKmer;

	i = 0;
	kmerDepthP = kmerDepth;
	for(l=0;l<kmerP->nUKmer;++l){
		ntmp = *kmerDepthP;
		if(ntmp < 1800) uniqKmerP[l].nhits = ntmp;  
		else uniqKmerP[l].nhits = 1800;
		uniqKmerP[l].kmerI = i;
		uniqKmerP[l].reptKmer = 0; 
		if(ntmp>=3 && ntmp<250){
			for(k=i;k<i+*kmerDepthP;++k){
				if(readInfP[fKmerSrt[k]->readId].kmerI < 0) readInfP[fKmerSrt[k]->readId].kmerI = l;
			}
			totDepth+= (long) ntmp;
			totUsedKmer++;
		}
		i += *kmerDepthP++;
	}
	kmerP->totUsedKmer = totUsedKmer;
	kmerP->totDepth = totDepth;
	free(kmerDepth);
}

int intCmp(const void *v1, const void *v2)
{
	return(*(int *)v1 - *(int *)v2);
}

void lnkKmerLst(KmerDB *kmerP)
{
	int i, j;
	FullKmerL **fKmerSrt = kmerP->fKmerSrt;
	UniqKmerL *uniqKmerP = kmerP->uKmer;
	FullKmerL *fullKmerP = kmerP->fKmer;
	FullKmerL *ptr;
	int *lnkKmerP1, *lnkKmerP2, *lnkKmerP3;
	int size, nLnkKmer, nlnkKmerT;

	nlnkKmerT = 0;

	lnkKmerP1 = kmerP->lnkKmerL;
	for(i=0;i<kmerP->nUKmer;++i){
		nLnkKmer = 0;
		lnkKmerP2 = lnkKmerP1;
		for(j=0;j<uniqKmerP[i].nhits; ++j){
			ptr = fKmerSrt[uniqKmerP[i].kmerI + j];
			if(ptr->readId == (ptr+1)->readId && ptr != &fullKmerP[kmerP->nFKmer-1]){
				*lnkKmerP2++ = (ptr+1)->uKmerId;
				++nLnkKmer;
			}
			if(ptr->readId == (ptr-1)->readId && ptr != &fullKmerP[0]){
				*lnkKmerP2++ = (ptr-1)->uKmerId;
				++nLnkKmer;
			}
		}
		if(nLnkKmer == 0){
			uniqKmerP[i].lnkKmHead = NULL;
			uniqKmerP[i].lnkKmSize = 0;
			continue;
		}
		qsort(lnkKmerP1,nLnkKmer,sizeof(int),intCmp);
		size = 0;
		uniqKmerP[i].lnkKmHead = lnkKmerP1;
		lnkKmerP3 = lnkKmerP1;
		while(lnkKmerP3 != lnkKmerP2){
			*lnkKmerP1=*lnkKmerP3++;
			while(lnkKmerP3!=lnkKmerP2 && *lnkKmerP1==*lnkKmerP3) lnkKmerP3++;
			lnkKmerP1++;
			++size;
		}
		uniqKmerP[i].lnkKmSize = size;
		nlnkKmerT+=size;
	}
}

int shtPath(UniqKmerL *uniqKmerP, int preKmerI, int curKmerI)
{
	int pKmerI = preKmerI;
	int cKmerI = curKmerI;  
	int lnkKmSize, i, n;
	UniqKmerL *ptr;

	n = 0;
	while(n < 40){
		ptr = &uniqKmerP[cKmerI];
		lnkKmSize = ptr->lnkKmSize;
		if(lnkKmSize > 2) return(0); 
		if(lnkKmSize == 0) break; 
		if(lnkKmSize == 1){
			++n;
			if(*(ptr->lnkKmHead) == pKmerI) break; 
			pKmerI = cKmerI;
			cKmerI = *(ptr->lnkKmHead);
		}
		if(lnkKmSize == 2){
			++n;
			for(i=0;i<ptr->lnkKmSize;++i){
				if(*(ptr->lnkKmHead+i)!=pKmerI){
					pKmerI = cKmerI;
					cKmerI = *(ptr->lnkKmHead+i);
					break; 
				}
			}
		}
	}
	if(n>=15) return(0);
	else return(1);
}

void shtPathRmv(KmerDB *kmerP, Fastq *read, int qthresh)
{
	int i, j;
	FullKmerL **fKmerSrt = kmerP->fKmerSrt;
	UniqKmerL *uniqKmerP = kmerP->uKmer;
	int kmerL = kmerP->kmerL;
	int nKeep, nhits, pos, readId;
	UniqKmerL *uKmerP;
	FullKmerL *fKmerSrtP;
	int *keep = (int *) malloc(sizeof(int)*1800); 
	int *keeptr;
	int flag;
	keeptr = keep;

	for(i=0;i<kmerP->nUKmer;++i){
		nKeep = 0;
		keeptr = keep;
		for(j=0;j<uniqKmerP[i].lnkKmSize; ++j){
			int indexTmp = *(uniqKmerP[i].lnkKmHead+j);
			uKmerP = &uniqKmerP[indexTmp];
			nhits = uKmerP->nhits;
			if(nhits <= 0) continue;
			if(nhits == 1){
				fKmerSrtP = fKmerSrt[uKmerP->kmerI];
				readId = fKmerSrtP->readId;
				if(fKmerSrtP->rc == 1){
					pos = fKmerSrtP->readPos+kmerL-1;
				}
				else {
					pos = read[readId].length - 1 - fKmerSrtP->readPos;
				}
				if(read[readId].qual[pos] <= qthresh) continue; 
				if(uKmerP->lnkKmSize <= 0) continue;
				flag = shtPath(uniqKmerP, i, indexTmp);
				if(flag==1) continue;
			}
			if(nhits == 2){
				if(uKmerP->lnkKmSize <= 0) continue;
				flag = shtPath(uniqKmerP, i, indexTmp);
				if(flag==1) continue;
			}
			*keeptr++ = indexTmp;
			++nKeep;
		}
		uniqKmerP[i].lnkKmSize = nKeep;
		keeptr = keep;
		for(j=0;j<nKeep;++j) *(j+uniqKmerP[i].lnkKmHead) = *keeptr++;
	}
	free(keep);
}

int kmer2binCode(long *kmerIndex, long *kmerIndexR, char *bases, int kmerL, char *baseIndex)
{
        int k;
        long baseCode, kmerCode, kmerCodeR;
        char *base;

        kmerCode = 0;
	kmerCodeR = 0;
	base = bases;
        for(k=kmerL-1;k>=0;k--) {
		baseCode = (long) baseIndex[(int) *base];
		if(baseCode == 4L) return(kmerL-k);  
		kmerCode = kmerCode<<2;
		kmerCode += baseCode;
		baseCode = (long) baseIndex[(int) *base+1];
		baseCode <<= (kmerL-1-k)*2;
		kmerCodeR += baseCode;
                ++base;
        }
	*kmerIndex = kmerCode;
	*kmerIndexR = kmerCodeR;
	return(0);
}

int base2binCode(long *kmerIndex, long *kmerIndexR, char *bases, int kmerL, char *baseIndex, long mask)
{

        long baseCode, kmerCode, kmerCodeR;
        char *base;

        kmerCode = *kmerIndex & mask;
	kmerCodeR = *kmerIndexR>>2;
	base = bases+kmerL-1;
	baseCode = (long) baseIndex[(int) *base];
	if(baseCode == 4L) return(kmerL);  
	kmerCode = kmerCode<<2;
	kmerCode += baseCode;
	baseCode = (long) baseIndex[(int) *base+1];
	baseCode <<= (kmerL-1)*2;
	kmerCodeR += baseCode;
	*kmerIndex = kmerCode;
	*kmerIndexR = kmerCodeR;
	return(0);
}

void reverseString(char *string, long len)
{
        long half, i;
        char tmp;
        char *f, *b;

        half = len/2;
        f = string;
        b = string + len - 1;
        for(i=0;i<half;++i){
                tmp = *f;
                *f++ = *b;
                *b-- = tmp;
        }
}

void complementBase(char *base)
{
        switch( *base ){
                case 'A':
                        *base = 'T';
                        break;
                case 'C':
                        *base = 'G';
                        break;
                case 'G':
                        *base = 'C';
                        break;
                case 'T':
                        *base = 'A';
                        break;
                case 'a':
                        *base = 't';
                        break;
                case 'c':
                        *base = 'g';
                        break;
                case 'g':
                        *base = 'c';
                        break;
                case 't':
                        *base = 'a';
                        break;
                case 'N':
                case 'n':
                        break;
                default:
                        fprintf(stderr, "I dont know this letter %c ! unchanged\n",*base);
        }
}


void reverseComplement(char *bases, long len)
{
        char *base;
        int i;

        reverseString(bases,len);
        base = bases;
        for(i=0;i<len;++i) complementBase(base++);
}

void read2kmerL(SeqDB *seqP, KmerDB *kmerDB)
{
        long hashIndex=0, hashIndexR=0;
        int i, s, length;
        char *ptr;
        long nKmer;
	FullKmerL *tmpPtr;

	int kmerL = kmerDB->kmerL;
	int n=kmerDB->nFKmer;
	FullKmerL *kmerPtr = &(kmerDB->fKmer[n]);
	long mask = (1L<<(2*(kmerL-1)))-1;
	int m;
	int Nbase;
	int p;

	for(s=0; s<seqP->nRead; ++s){
		length = seqP->read[s].length;
	        nKmer = (length - kmerL + 1);
	        ptr = seqP->read[s].base;
		tmpPtr = kmerPtr;
		Nbase = 1;
		i = 0;
		m = 0;
		while(i<nKmer){        
			if(Nbase==1){
				p = kmer2binCode(&hashIndex, &hashIndexR, ptr, kmerL, seqP->baseIndex);
				if(p != 0){
					ptr+=p;
					i+=p;
					Nbase = 1;
					continue;
				}
				Nbase = 0;
			}
			else {
				p = base2binCode(&hashIndex, &hashIndexR, ptr, kmerL, seqP->baseIndex, mask);
				if(p != 0){
					ptr+=p;
					i+=p;
					Nbase = 1;
					continue;
				}
				Nbase = 0;
			}
	                if(hashIndex >= hashIndexR){
				kmerPtr->kmerC = hashIndexR;
				kmerPtr->readPos = length-(i+kmerL);
				kmerPtr->rc = -1;
			}
			else {
				kmerPtr->kmerC = hashIndex;
				kmerPtr->readPos = i;
				kmerPtr->rc = 1;
			}
			if(hashIndex == hashIndexR) kmerPtr->kmerC = -100;
			kmerPtr->readId = s;
			++kmerPtr;
	                ++ptr;
			++m;
			++i;
	        }
		n+=m;
	}
	kmerDB->nFKmer += n;
}

