/**************************************************************************
 *
 *-------------------------------------------------------------------------
 * Project : Short Read Assembler 
 *
 *-------------------------------------------------------------------------
 #######################################################################
 # This software has been created by Genome Research Limited (GRL).    #
 # GRL hereby grants permission to use, copy, modify and distribute    #
 # this software and its documentation for non-commercial purposes     #
 # without fee at the user's own risk on the basis set out below.      #
 # GRL neither undertakes nor accepts any duty whether contractual or  #
 # otherwise in connection with the software, its use or the use of    #
 # any derivative, and makes no representations or warranties, express #
 # or implied, concerning the software, its suitability, fitness for   #
 # a particular purpose or non-infringement.                           #
 # In no event shall the authors of the software or GRL be responsible #
 # or liable for any loss or damage whatsoever arising in any way      #
 # directly or indirectly out of the use of this software or its       #
 # derivatives, even if advised of the possibility of such damage.     #
 # Our software can be freely distributed under the conditions set out #
 # above, and must contain this copyright notice.                      #
 #######################################################################
 *
 *  Author : Yong Gu (ygu@sanger.ac.uk) 
 *
 *  Copyright (C) 2009 by Genome Research Limited, All rights reserved.
 *
 **************************************************************************/

#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <dirent.h>
#include <values.h>
#include <string.h>
#include <math.h>
#include <unistd.h>

#include "utils.h"
#include "fpathM.h"
#include "kmerP.h"
#include "kmerW.h"

int initExtDir(FullKmerL **fKmerSrtH, int nHits, int kmerL, char *cnsSeq, Fastq *read) 
{
	int i;
	int backward, forward;
	int extDir;
	int readId;

	backward = 0; forward=0;
	for(i=0;i<nHits;++i){
		readId = fKmerSrtH[i]->readId;
		if(fKmerSrtH[i]->rc == -1){
			if(readId == (fKmerSrtH[i]-1)->readId){
				++backward;
			}
		}
		else {
			if(readId == (fKmerSrtH[i]+1)->readId){
				++forward;
			}
		}
	}
	if(forward==0 && backward==0) return(0);
	if(forward >= backward) extDir = 1;
	else extDir = -1;
	return(extDir);
}


void ctgOffset(ExtCntlD *extCntlP, int readId, int pairedId, int nHits, ReadInf *readInfP)
{
	if(extCntlP->numCon > 100){
		if(readInfP[pairedId].ctgId == extCntlP->nContig && readInfP[pairedId].ctgOffset>0 && extCntlP->numCon-readInfP[pairedId].ctgOffset>30 && extCntlP->numCon-readInfP[pairedId].ctgOffset<extCntlP->maxExtLen && nHits<extCntlP->numSetCover){
			readInfP[readId].ctgId = extCntlP->nContig;   
			readInfP[readId].ctgOffset = extCntlP->numCon+1; 
		}
	}
	if(nHits<extCntlP->numSetCover){ 
		readInfP[readId].ctgId = extCntlP->nContig; 
		if(readInfP[readId].ctgOffset == 0)
			readInfP[readId].ctgOffset = extCntlP->numCon+1;
	}
}

void ctgOffsetE(ExtCntlD *extCntlP, int readId, int pairedId, int nHits, ReadInf *readInfP)
{
		readInfP[readId].ctgId = extCntlP->nContig; 
		readInfP[readId].ctgOffset = extCntlP->numCon+1;
}

void gapStats(ExtCntlD *extCntlP, BaseCount *baseCount, int kmerL, int pairedId, int baseI, ReadInf *readInfP, int qual)
{
	int gap;

	gap = abs(extCntlP->insertSize - abs(extCntlP->numCon-readInfP[pairedId].ctgOffset)-kmerL-1);

	if(gap < 100 && readInfP[pairedId].ctgId == extCntlP->nContig && readInfP[pairedId].ctgOffset>0){
		baseCount[baseI].gapSize += gap;
		baseCount[baseI].nGap++;
		baseCount[baseI].qual = qual;
	}
}


void baseCnt(ExtCntlD *extCntlP,int nHits, FullKmerL **fKmerSrtH, BaseCount *baseCount, Fastq *read, int kmerL, int extDir, char *baseIndex, ReadInf *readInfP)
{
	int i;
	int readId, rc, pos, baseI;
	short readPos;
	FullKmerL *ptr;
	int pairedId;
	int id_offset;

	for(i=0;i<nHits;++i){
		readId = fKmerSrtH[i]->readId;
		readPos = fKmerSrtH[i]->readPos;
		pairedId = readInfP[readId].pairedId;
		ctgOffset(extCntlP, readId, pairedId, nHits, readInfP);
	}
	extCntlP->num_454negs = 0;
	for(i=0;i<nHits;++i){
		readId = fKmerSrtH[i]->readId;
		readPos = fKmerSrtH[i]->readPos;
		pairedId = readInfP[readId].pairedId;
#ifdef SMALLSET
#else
		extCntlP->readList[extCntlP->readCnt++] = readId;
#endif
		rc = fKmerSrtH[i]->rc;
		pos = -1;
		if(extDir*rc == 1){
			ptr = fKmerSrtH[i] + 1;   
			if(readId != ptr->readId){
				 continue;
			}
			if(ptr->kmerC == -100) continue;
			if(rc == -1) pos = read[readId].length - readPos;
			else pos = readPos + kmerL;
			baseI = baseIndex[(int)read[readId].base[pos]];
		}
		else {
			ptr = fKmerSrtH[i] - 1;  
			if(readId != ptr->readId){
				continue;
			}
			if(ptr->kmerC == -100) continue;
			if(rc == -1) pos = read[readId].length - readPos - kmerL - 1;
			else pos = readPos - 1;
			baseI = baseIndex[(int)read[readId].base[pos]+1];
		}
		if(pos != -1){ 

			if(fKmerSrtH[i]->kmerC != -100)gapStats(extCntlP, baseCount, kmerL, pairedId, baseI, readInfP, read[readId].qual[pos]);

			baseCount[baseI].c++;   
			baseCount[baseI].uKmerId = ptr->uKmerId;
			if(readInfP[readId].postR == -2) readInfP[readId].postR = extCntlP->numCon;
			if(extCntlP->numCon-readInfP[readId].postR > read[readId].length){
					readInfP[readId].postR = extCntlP->numCon;
					readInfP[readId].post = 0;
			}
			readInfP[readId].post++;
		
			readInfP[readId].postR = extCntlP->numCon;
			if(readInfP[readId].post > baseCount[baseI].maxOffset){
				baseCount[baseI].maxOffset = readInfP[readId].post;
			}
			if(extDir*rc==1) id_offset = pos+1;
 			else id_offset = read[readId].length - pos - 1;
  			if(readInfP[readId].post == extCntlP->numCon && extCntlP->numCon >= 20){
				baseCount[baseI].num_454cuts++;
			}
			if(read[readId].qual[pos] < extCntlP->qThresh){
				baseCount[baseI+5].c++;  
				baseI = baseIndex['N'];
				baseCount[baseI].c++;
			}
		}
	}
}

void baseCntE(ExtCntlD *extCntlP,int nHits, FullKmerL **fKmerSrtH, BaseCount *baseCount, Fastq *read, int kmerL, int extDir, char *baseIndex, ReadInf *readInfP)
{
	int i;
	int readId, rc, pos, baseI;
	short readPos;
	FullKmerL *ptr;
	int pairedId;

	for(i=0;i<nHits;++i){
		readId = fKmerSrtH[i]->readId;
		readPos = fKmerSrtH[i]->readPos;
		pairedId = readInfP[readId].pairedId;
		ctgOffsetE(extCntlP, readId, pairedId, nHits, readInfP);
		rc = fKmerSrtH[i]->rc;
		pos = -1;
		if(extDir*rc == 1){
			ptr = fKmerSrtH[i] + 1;   
			if(readId != ptr->readId) continue;
			if(rc == -1) pos = read[readId].length - readPos;
			else pos = readPos + kmerL;
			baseI = baseIndex[(int)read[readId].base[pos]];
		}
		else {
			ptr = fKmerSrtH[i] - 1;  
			if(readId != ptr->readId) continue;
			if(rc == -1) pos = read[readId].length - readPos - kmerL - 1;
			else pos = readPos - 1;
			baseI = baseIndex[(int)read[readId].base[pos]+1];
		}
		if(pos != -1){ 

			gapStats(extCntlP, baseCount, kmerL, pairedId, baseI, readInfP, read[readId].qual[pos]);

			baseCount[baseI].c++;   
			baseCount[baseI].uKmerId = ptr->uKmerId;
			readInfP[readId].post++;
			if(readInfP[readId].post > baseCount[baseI].maxOffset){
				baseCount[baseI].maxOffset = readInfP[readId].post;
			}
			if(read[readId].qual[pos] < extCntlP->qThresh){
				baseCount[baseI+5].c++;  
				baseI = baseIndex['N'];
				baseCount[baseI].c++;
			}
		}
	}
}


int nGapCmp(const void *v1, const void *v2)
{
	return((*(BaseCount **)v2)->nGap - (*(BaseCount **)v1)->nGap);
}

int gapSizeCmp(const void *v1, const void *v2)
{
	return((*(BaseCount **)v1)->gapSize - (*(BaseCount **)v2)->gapSize);
}

int getNPaired(BaseCount *baseCount)
{
	int i, nPaired;

	nPaired = 0;
	for(i=0;i<4;++i){
		nPaired += baseCount[i].nGap;
	}
	return(nPaired);
}

int pairCheck(ExtCntlD *extCntlP, BaseCount *baseCount)
{
	int i;
	int nPaired = 0;
	BaseCount *baseCountGap[5], *baseCountGap2[5];

	for(i=0;i<4;++i){
		baseCountGap[i] = &baseCount[i];
		baseCountGap2[i] = &baseCount[i];
		if(baseCountGap[i]->nGap!=0 && baseCountGap[i]->nGap!=1)
			baseCountGap[i]->gapSize /= baseCountGap[i]->nGap;
		else baseCountGap[i]->gapSize = 99999999;
		nPaired += baseCountGap[i]->nGap;
	}
	if(nPaired == 0) return(-2);

	qsort(baseCountGap,4,sizeof(BaseCount *),nGapCmp);
	if(baseCountGap[1]->nGap==0) return(baseCountGap[0]->uKmerId);
	if(baseCountGap[1]->nGap>=2){
		qsort(baseCountGap2,4,sizeof(BaseCount *),gapSizeCmp);  
		if(baseCountGap2[1]->nGap == 0) return(baseCountGap2[0]->uKmerId); 
		if(baseCountGap2[0]->gapSize < 80 && baseCountGap2[1]->gapSize < 80){ 
			if(baseCountGap2[0]->nGap >= baseCountGap2[1]->nGap && baseCountGap2[0]->gapSize < baseCountGap2[1]->gapSize)  	
				return(baseCountGap2[0]->uKmerId);
			else
				return(-1);

		}
		else {
			if(baseCountGap2[0]->gapSize < baseCountGap2[1]->gapSize)
				return(baseCountGap2[0]->uKmerId);
			else return(-1); 
		}
	}
	else {
		if(baseCountGap[1]->nGap==1 && nPaired >=3){
			if(baseCountGap[0]->nGap>=2){  
				if(baseCountGap[1]->qual < 20) return(baseCountGap[0]->uKmerId);
				else return(-1);
			}
		}
		else
			return(-1);
	}
	return(0);
}

int baseCCmp(const void *v1, const void *v2)
{
	return((*(BaseCount **)v2)->c - (*(BaseCount **)v1)->c);
}

int checkBreak(int nHits, BaseCount **baseCountPL)
{
	int breakup = 1;

	int senseFlag = 1;
	if(senseFlag == 0){
		if((baseCountPL[0]->c > 0 && 1.0*baseCountPL[1]->c/baseCountPL[0]->c < 0.1) || (baseCountPL[0]->c > 3 && baseCountPL[1]->c == 1)){
			breakup = 0;
		}
	}
	else {
		if(baseCountPL[0]->c > 0 && baseCountPL[1]->c==0){
			breakup = 0;
		}
	}
	if(baseCountPL[0]->c == 1 && baseCountPL[1]->c == 0){
		breakup = 0;
	}

	float rate = 1.0*baseCountPL[0]->c/nHits;
	if(baseCountPL[1]->c == 0) rate = 2.0;
	if(breakup == 1) return(1);
	
	if(baseCountPL[0]->c == 0){
		return(1); 
	}
	return(0);
}

int checkShtPath(UniqKmerL *uniqKmerP, int uniqKmerI, int nHits, int maxMove2)
{
	int shtFlag = 0;
	int nLong = 0;
	int lnkKmSize;
	int curKmerI;
	int newMaxMove2;
	int i;

	newMaxMove2 = maxMove2;
	
	lnkKmSize = uniqKmerP[uniqKmerI].lnkKmSize;

	if(lnkKmSize == 2) return(1); 

	for(i=0;i<lnkKmSize;++i){
		curKmerI = *(uniqKmerP[uniqKmerI].lnkKmHead + i); 
		shtFlag = shtPath(uniqKmerP, uniqKmerI, curKmerI);
		if(shtFlag == 0) ++nLong;
	}


	if(uniqKmerP[uniqKmerI].lnkKmSize==3 && nLong == 2) return(1);
	return(0); 
} 

int ppatchCheck(BaseCount **baseCountPL)
{
	int i;
	int num_ppatch = 0;
	int map_post[4], map_kmer[4];

	for(i=0;i<4;++i){
		if(baseCountPL[i]->c>=2){
			map_post[num_ppatch] = baseCountPL[i]->maxOffset;
			map_kmer[num_ppatch] = baseCountPL[i]->uKmerId;
			num_ppatch++;
		}
	}
	int set_offset = 4; 
	if(num_ppatch == 2){
		if(map_post[0] < set_offset && map_post[1] > set_offset) return(map_kmer[1]);
		if(map_post[1] < set_offset && map_post[0] > set_offset) return(map_kmer[0]);
	}
	return(-1);
}

int nextKmer(int extDir, SeqDB *seqP, UniqKmerL *uniqKmerP, FullKmerL **fKmerSrtH, ExtCntlD *extCntlP, int kmerL, char *cnsBase, int *cnsHit, ReptKmerId *reptKmerIdP, int uniqKmerI, ReadInf *readInfP)
{
	int i;
	char *baseIndex = seqP->baseIndex;
	BaseCount baseCount[10], *baseCountPL[5], *baseCountPL2[5];
	BaseCount baseCount2[5];
	Fastq *read = seqP->read;
	int maxMove2;
	float rate=1.0;

	int nHits = uniqKmerP[uniqKmerI].nhits;

	for(i=0;i<5;++i){
		baseCount[i].c = 0;
		baseCount[i].b = baseIndex[i];
		baseCount[i].maxOffset = 0;
		baseCount[i].num_454cuts = 0;
		baseCount[i].nGap = 0;
		baseCount[i].qual = 0;
		baseCount[i].gapSize = 0;
		baseCountPL[i] = &baseCount[i];
		baseCount[i+5].c = 0; 
		baseCount[i+5].b = baseIndex[i]; 
	}

	baseCnt(extCntlP, nHits, fKmerSrtH, baseCount, read, kmerL, extDir, baseIndex, readInfP);
	qsort(baseCountPL,4,sizeof(BaseCount *),baseCCmp);
	*cnsBase = baseCountPL[0]->b;
	*cnsHit = 0;
	for(i=0;i<4;++i) {
		*cnsHit += baseCountPL[i]->c;
	}

	if(baseCountPL[0]->c == 0) {
		return(-1);
	}
	int eHits = 0;
	for(i=0;i<4;++i) eHits += baseCountPL[i]->c;

	if(baseCountPL[1]->c == 0 && baseCountPL[0]->c >= 1) return(baseCountPL[0]->uKmerId);
	for(i=0;i<5;++i){
		baseCount2[i].c = baseCount[i].c - baseCount[i+5].c; 
		baseCount2[i].uKmerId = baseCount[i].uKmerId;
		baseCountPL2[i] = &baseCount2[i];
	}
	qsort(baseCountPL2,4,sizeof(BaseCount *),baseCCmp);


	maxMove2 = baseCountPL[1]->c;
	if(baseCountPL[1]->c >= 2 && uniqKmerP[uniqKmerI].lnkKmSize>=3){
		if(checkShtPath(uniqKmerP, uniqKmerI, nHits, baseCountPL[1]->c)){
			maxMove2 = 0;
		}
		else { 
			int breakFlag = 0;
			if((baseCountPL2[1]->c==2) && uniqKmerP[baseCountPL2[1]->uKmerId].nhits >= 4) breakFlag = 1;
			if(baseCountPL2[1]->c>=3) breakFlag = 1;
			if(breakFlag==1 && extCntlP->pairCheck==0){  
				int pbreak;
				pbreak = ppatchCheck(baseCountPL);
				if(pbreak==-1){
					rate = 1.0*baseCountPL[0]->c;
					if(maxMove2 == 0) rate = 2.0;
					else rate /= maxMove2;
					if(rate>1.4 && maxMove2<extCntlP->misMatch) {
					}
					else {
						for(i=0;i<4;++i){
							if(baseCountPL[i]->c<2) break;
							if(uniqKmerP[uniqKmerI].reptKmer == 1) continue;
							reptKmerIdP->uKmerId[reptKmerIdP->nRept].word = uniqKmerI;
							reptKmerIdP->uKmerId[reptKmerIdP->nRept++].next = baseCountPL[i]->uKmerId;
						}
						uniqKmerP[uniqKmerI].reptKmer = 1;
					}
				}
				if(eHits > 1) return(-2);
				else return(-1);
			}
		} 
#ifndef RELEASE
		if(maxMove2==0&&baseCountPL[0]->c==baseCountPL[1]->c){
			if(eHits > 1) return(-2);
			else return(-1);
		}
#endif
	}
	if(eHits == 3 && baseCountPL[0]->c==2){
		if(baseCountPL[4]->c==0) {
			if(eHits > 1) return(-2);
			else return(-1);
		}
	}

	float pathRate = 1.0;
	rate = 1.0*maxMove2;
	rate = rate/baseCountPL[0]->c;
	pathRate = 1.0*baseCountPL[0]->c;
	pathRate = pathRate/eHits;
	if(!(pathRate<0.65 || rate>0.52 || maxMove2>=extCntlP->misMatch)) return(baseCountPL[0]->uKmerId);

	int vbreak;
	vbreak = checkBreak(nHits, baseCountPL2);  
	if(vbreak == 0){    
		return(baseCountPL2[0]->uKmerId);
	}

	if(extCntlP->fuzzyCheck == 1 && baseCountPL[0]->c<=2){
			if(baseCountPL2[1]->c>0) return(baseCountPL[0]->uKmerId);
	}

	int Break_454 = 1;
	int nPaired = 0;
	int num_454pats=0;
	if(extCntlP->pairCheck == 1 && extCntlP->fuzzyCheck == 0){
		nPaired = getNPaired(baseCount);
	}
	if(nPaired == 0){	
		for(i=0;i<4;++i){
			if(baseCountPL[i]->c>=2){
				if(baseCountPL[i]->num_454cuts>0){
					num_454pats++; 
				}
			}
		}
		Break_454 = 1;
		if(num_454pats == 1){
			Break_454 = 0; 
		}
	}
	if(Break_454 == 0 && extCntlP->numCon < extCntlP->maxExtLen){
		return(0); 
	}

	if(nPaired == 0 && extCntlP->fuzzyCheck == 0){
		int pbreak;
		pbreak = ppatchCheck(baseCountPL);
		if(pbreak>=0)return(pbreak);

		rate = 1.0*baseCountPL[0]->c;
		if(maxMove2 == 0) rate = 2.0;
		else rate /= maxMove2;
		if((rate>1.4 && maxMove2<extCntlP->misMatch)||vbreak==0) return(baseCountPL[0]->uKmerId); 
		else {
			for(i=0;i<4;++i){
				if(baseCountPL[i]->c<2) break;
				if(uniqKmerP[uniqKmerI].reptKmer == 1) continue;
				reptKmerIdP->uKmerId[reptKmerIdP->nRept].word = uniqKmerI;
				reptKmerIdP->uKmerId[reptKmerIdP->nRept++].next = baseCountPL[i]->uKmerId;
			}
			uniqKmerP[uniqKmerI].reptKmer = 1;
			if(eHits > 1) return(-2);
			else return(-1);
		}
	}
	else if(extCntlP->fuzzyCheck == 0) {
		int pairBreak;
		pairBreak = pairCheck(extCntlP, baseCount);
		if(pairBreak >= 0) return(pairBreak);

		if(eHits > 1) return(-2);
		else return(-1);
	}
	if(vbreak == 0) return(baseCountPL[0]->uKmerId);
	else {
		if(eHits > 1) return(-2);
		else return(-1);
	}
}

int nextKmerE(int extDir, SeqDB *seqP, UniqKmerL *uniqKmerP, FullKmerL **fKmerSrtH, ExtCntlD *extCntlP, int kmerL, char *cnsBase, int *cnsHit, ReptKmerId *reptKmerIdP, int uniqKmerI, ReadInf *readInfP)
{
	int i;
	char *baseIndex = seqP->baseIndex;
	BaseCount baseCount[10], *baseCountPL[5], *baseCountPL2[5];
	BaseCount baseCount2[5];
	Fastq *read = seqP->read;
	int maxMove2;

	int nHits = uniqKmerP[uniqKmerI].nhits;

	for(i=0;i<5;++i){
		baseCount[i].c = 0;
		baseCount[i].b = baseIndex[i];
		baseCount[i].maxOffset = 0;
		baseCount[i].nGap = 0;
		baseCount[i].gapSize = 0;
		baseCountPL[i] = &baseCount[i];
		baseCount[i+5].c = 0; 
		baseCount[i+5].b = baseIndex[i]; 
	}

	baseCntE(extCntlP, nHits, fKmerSrtH, baseCount, read, kmerL, extDir, baseIndex, readInfP);
	qsort(baseCountPL,4,sizeof(BaseCount *),baseCCmp);
	*cnsBase = baseCountPL[0]->b;
	*cnsHit = 0;
	for(i=0;i<4;++i) {
		*cnsHit += baseCountPL[i]->c;
	}

	if(baseCountPL[0]->c == 0) {
		return(-1);
	}
	int eHits = 0;
	for(i=0;i<4;++i) eHits += baseCountPL[i]->c;

	for(i=0;i<5;++i){
		baseCount2[i].c = baseCount[i].c - baseCount[i+5].c; 
		baseCount2[i].uKmerId = baseCount[i].uKmerId;
		baseCountPL2[i] = &baseCount2[i];
	}
	qsort(baseCountPL2,4,sizeof(BaseCount *),baseCCmp);

	maxMove2 = baseCountPL[1]->c;
	if(baseCountPL[1]->c >= 2){
		int breakFlag = 0;
		if(baseCountPL2[1]->c==1) baseCountPL2[1]->c = 0;
		if((baseCountPL2[1]->c==1 || baseCountPL2[1]->c==2) && uniqKmerP[baseCountPL2[1]->uKmerId].nhits >= 4) breakFlag = 1;
		if(baseCountPL2[1]->c>=3) breakFlag = 1;

		if(breakFlag==1){     
			return(-1); 
		}
	}
	if(maxMove2==0&&baseCountPL[0]->c==baseCountPL[1]->c){
		return(-1);
	}
	if(((1.0*baseCountPL[0]->c/eHits)<0.65 || (1.0*maxMove2/baseCountPL[0]->c)>0.52 || maxMove2>=extCntlP->misMatch)) {
		int nPaired;
		nPaired = getNPaired(baseCount);
	
		float rate = 1.0*baseCountPL[0]->c;
		if(maxMove2 == 0) rate = 2.0;
		else rate /= maxMove2;
		if(rate>1.4 && maxMove2<extCntlP->misMatch){
		}
		else {
			return(-1); 
		}
	}
	extCntlP->breakKmer = baseCountPL[0]->uKmerId;
	if(extCntlP->numCon < 4 && baseCountPL[0]->uKmerId==extCntlP->reptWord){
		extCntlP->breakKmer = extCntlP->reptWord;
		return(-1);
	}
	return(baseCountPL[0]->uKmerId);
}
 
int nextExtDir(int uKmerId, int extDir, int nHits, FullKmerL **fKmerSrtH)  
{
	int i;
	FullKmerL *ptr;
	int readId, rc;
	int nxtExtDir;

	nxtExtDir = extDir;
	for(i=0;i<nHits;++i){
		readId = fKmerSrtH[i]->readId;
		rc = fKmerSrtH[i]->rc;

		if(extDir*rc == -1) ptr = fKmerSrtH[i] - 1;
		else ptr = fKmerSrtH[i] + 1;

		if(readId != ptr->readId) continue;
		if(ptr->uKmerId != uKmerId) continue;
		if(ptr->rc != rc) nxtExtDir = -extDir;
		break;
	}
	return(nxtExtDir);
}

int kmerWalking(char *cnsSeq, int initUniqKmerI, KmerDB *kmerP, SeqDB *seqP, ReadInf *readInfP, ReptKmerId *reptKmerIdP, ExtCntlD *extCntlP, int initDir)
{
	FullKmerL **fKmerSrt = kmerP->fKmerSrt;
	UniqKmerL *uniqKmerP = kmerP->uKmer;
	int kmerL = kmerP->kmerL;
	ReadInf *readInfPtr;
	int numCns, extDir;
	int i;
	int uniqKmerI=initUniqKmerI;
	char *cnsSeqP = cnsSeq;

	extDir = initDir;
	numCns = 0; 
	int sglHit=0, cnsHit;
	if(uniqKmerP[uniqKmerI].nhits <= 1) ++sglHit;
	else sglHit = 0;

	extCntlP->readCnt = 0;

	while(1){ 
		if(uniqKmerP[uniqKmerI].nhits == 0 || uniqKmerP[uniqKmerI].nhits > extCntlP->depth) break;
		if(fKmerSrt[uniqKmerP[uniqKmerI].kmerI]->kmerC == -100) break;
		int uniqKmerI_old = uniqKmerI;
		int extDir_old = extDir;
		extCntlP->numCon=numCns;
		uniqKmerI = nextKmer(extDir, seqP, uniqKmerP, &fKmerSrt[uniqKmerP[uniqKmerI].kmerI], extCntlP, kmerL, cnsSeqP, &cnsHit, reptKmerIdP, uniqKmerI, readInfP);
		kmer2base(fKmerSrt[uniqKmerP[uniqKmerI_old].kmerI]->kmerC, cnsSeqP, kmerL, extDir_old, seqP->baseIndex);
 		if(uniqKmerI==-2){
			kmer2seq(fKmerSrt[uniqKmerP[uniqKmerI_old].kmerI]->kmerC, cnsSeqP, kmerL, extDir_old, seqP->baseIndex);
			cnsSeqP+=kmerL; numCns+=kmerL;
		}
		else {
			++cnsSeqP; ++numCns;
		}
		if(uniqKmerI==-1 || uniqKmerI==-2) break;
		if(cnsHit <= 1) ++sglHit;
		else sglHit = 0;
		if(numCns >= extCntlP->maxExtLen) break;
		if(sglHit >= extCntlP->sglHitThr) break;
		extDir = nextExtDir(uniqKmerI, extDir, uniqKmerP[uniqKmerI_old].nhits, &fKmerSrt[uniqKmerP[uniqKmerI_old].kmerI]);
	}
#ifdef SMALLSET
	for(i=0;i<seqP->nRead;++i){
		readInfPtr = &readInfP[i];
		readInfPtr->post = 0;
		readInfPtr->postR = -2;
		readInfPtr->ctgId = -1;
		readInfPtr->ctgOffset = 0;
	}
#else
	for(i=0;i<extCntlP->readCnt;++i){
		readInfPtr = &readInfP[extCntlP->readList[i]];
		readInfPtr->post = 0;
		readInfPtr->postR = -2;
		readInfPtr->ctgId = -1;
		readInfPtr++->ctgOffset = 0;
	}
#endif

#ifdef RELEASE
	if(numCns <= 1) return(0);
#endif
	if(numCns > 3000) {
		cnsSeqP -= numCns-3000;
		numCns = 3000;
	}
	*cnsSeqP++ = '\0';
	return(numCns);
}


int kmerWalkingE(char *cnsSeq, int initUniqKmerI, KmerDB *kmerP, SeqDB *seqP, ReadInf *readInfP, ReptKmerId *reptKmerIdP, ExtCntlD *extCntlP, int initDir)
{
	FullKmerL **fKmerSrt = kmerP->fKmerSrt;
	UniqKmerL *uniqKmerP = kmerP->uKmer;
	int kmerL = kmerP->kmerL;
	int numCns, extDir;
	char *cnsSeqP = cnsSeq;
	int uniqKmerI = initUniqKmerI;

	extDir = initDir;
	int initExtDir = extDir;
	numCns = 0; 
	int sglHit=0, cnsHit;
	if(uniqKmerP[uniqKmerI].nhits <= 1) ++sglHit;
	else sglHit = 0;
	while(1){ 
		if(uniqKmerP[uniqKmerI].nhits < 2 || uniqKmerP[uniqKmerI].nhits > extCntlP->depth) break;
		if(fKmerSrt[uniqKmerP[uniqKmerI].kmerI]->kmerC == -100) break;
		int uniqKmerI_old = uniqKmerI;
		int extDir_old = extDir;
		extCntlP->numCon=numCns;
		uniqKmerI = nextKmerE(extDir, seqP, uniqKmerP, &fKmerSrt[uniqKmerP[uniqKmerI].kmerI], extCntlP, kmerL, cnsSeqP, &cnsHit, reptKmerIdP, uniqKmerI, readInfP);
		kmer2base(fKmerSrt[uniqKmerP[uniqKmerI_old].kmerI]->kmerC, cnsSeqP, kmerL, extDir_old, seqP->baseIndex);
 		if(uniqKmerI==-2){
			kmer2seq(fKmerSrt[uniqKmerP[uniqKmerI_old].kmerI]->kmerC, cnsSeqP, kmerL, extDir_old, seqP->baseIndex);
			cnsSeqP+=kmerL; numCns+=kmerL;
		}
		else {
			++cnsSeqP; ++numCns;
		}
		if(uniqKmerI==-1 || uniqKmerI==-2) break;
		if(cnsHit <= 1) ++sglHit;
		else sglHit = 0;
		if(numCns >= 700) break;
		if(sglHit >= extCntlP->sglHitThr) break;
		extDir = nextExtDir(uniqKmerI, extDir, uniqKmerP[uniqKmerI_old].nhits, &fKmerSrt[uniqKmerP[uniqKmerI_old].kmerI]);
	}
	if(initExtDir==-1) initExtDir = 0;
	if(numCns < 1) {
		return(0);
	}
	*cnsSeqP++ = '\0';
	return(numCns);
}

