/**************************************************************************
 *
 *-------------------------------------------------------------------------
 * Project : Short Read Assembler 
 *
 *-------------------------------------------------------------------------
 #######################################################################
 # This software has been created by Genome Research Limited (GRL).    #
 # GRL hereby grants permission to use, copy, modify and distribute    #
 # this software and its documentation for non-commercial purposes     #
 # without fee at the user's own risk on the basis set out below.      #
 # GRL neither undertakes nor accepts any duty whether contractual or  #
 # otherwise in connection with the software, its use or the use of    #
 # any derivative, and makes no representations or warranties, express #
 # or implied, concerning the software, its suitability, fitness for   #
 # a particular purpose or non-infringement.                           #
 # In no event shall the authors of the software or GRL be responsible #
 # or liable for any loss or damage whatsoever arising in any way      #
 # directly or indirectly out of the use of this software or its       #
 # derivatives, even if advised of the possibility of such damage.     #
 # Our software can be freely distributed under the conditions set out #
 # above, and must contain this copyright notice.                      #
 #######################################################################
 *
 *  Author : Yong Gu (ygu@sanger.ac.uk) 
 *
 *  Copyright (C) 2009 by Genome Research Limited, All rights reserved.
 *
 **************************************************************************/

#include <stdio.h>
#include "fpathM.h"
#include "kmerP.h"

#ifndef KMERW_H
#define KMERW_H

typedef struct _readInf {
	int kmerI;
	int ctgId;  
	int ctgOffset; 
	int pairedId;   
	int post;	// change into short
	int postR;	
} ReadInf;

typedef struct _baseCount {
	int c;	
	char b;	
	int nhits;
	int uKmerId;
	int maxOffset;
	int gapSize;
	int nGap;
	int qual; 	
	int num_454cuts;
} BaseCount;

typedef struct _reptKmer {
	int word;
	int next;
} ReptKmer;

typedef struct _reptKmerId {
	ReptKmer *uKmerId;
	int nRept;
} ReptKmerId;

typedef struct _pileupD {
	long nextKmerC;
} PileupD;

typedef struct _resSeq {
	char *extSeqD;
	int nSeq;
	char **extSeq;
	int *sLen;
} ResSeq;

typedef struct _extCntlD {
	int fuzzyCheck;
	int pairCheck;
	int depth;
	int maxExtLen;
	int kmerL;
	int insertSize; 

	FILE *fpOut;
	int nExtRead;
	int maxRept;
	int maxExtRead;
	int qThresh;
	int sglHitThr; 
	int useAllRead;

	int misMatch;
	int numSetCover;
	int nContig;
	int numCon;
	int reptWord;
	int breakKmer;
	char *cnsSeq;     
	char *xcnsSeq;

	int num_454negs;
	int nextKmer;
	int eHits;

	int *readList;
	int readCnt;
	int *uniqKmerIList;
	int uniqKmerICnt;

} ExtCntlD;

void sortKmerL(KmerDB *kmerP, ReadInf *readInfP, int nRead);
int kmerWalking(char *cnsSeq, int initUniqKmerI, KmerDB *kmerP, SeqDB *seqP, ReadInf *readInfP, ReptKmerId *reptKmerIdP, ExtCntlD *extCntlP, int initDir);
int kmerWalkingE(char *cnsSeq, int initUniqKmerI, KmerDB *kmerP, SeqDB *seqP, ReadInf *readInfP, ReptKmerId *reptKmerIdP, ExtCntlD *extCntlP, int initDir);

#endif  /* !defined(KMERW_H) */
