/**************************************************************************
 *
 *-------------------------------------------------------------------------
 * Project : Short Read Assembler 
 *
 *-------------------------------------------------------------------------
 #######################################################################
 # This software has been created by Genome Research Limited (GRL).    #
 # GRL hereby grants permission to use, copy, modify and distribute    #
 # this software and its documentation for non-commercial purposes     #
 # without fee at the user's own risk on the basis set out below.      #
 # GRL neither undertakes nor accepts any duty whether contractual or  #
 # otherwise in connection with the software, its use or the use of    #
 # any derivative, and makes no representations or warranties, express #
 # or implied, concerning the software, its suitability, fitness for   #
 # a particular purpose or non-infringement.                           #
 # In no event shall the authors of the software or GRL be responsible #
 # or liable for any loss or damage whatsoever arising in any way      #
 # directly or indirectly out of the use of this software or its       #
 # derivatives, even if advised of the possibility of such damage.     #
 # Our software can be freely distributed under the conditions set out #
 # above, and must contain this copyright notice.                      #
 #######################################################################
 *
 *  Author : Yong Gu (ygu@sanger.ac.uk) 
 *
 *  Copyright (C) 2009 by Genome Research Limited, All rights reserved.
 *
 **************************************************************************/

#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <dirent.h>
#include <values.h>
#include <string.h>
#include <math.h>
#include <unistd.h>

#include "utils.h"
#include "fpathM.h"
#include "kmerP.h"
#include "kmerW.h"

void initialising(SeqDB *seqP)
{
	int i;
	char *baseIndex=seqP->baseIndex;

	for(i=0;i<256;++i) baseIndex[i] = -1;
	baseIndex['A'] = 0; baseIndex[0] = 'A';
	baseIndex['C'] = 1; baseIndex[1] = 'C';
	baseIndex['G'] = 2; baseIndex[2] = 'G';
	baseIndex['T'] = 3; baseIndex[3] = 'T';
	baseIndex['N'] = 4; baseIndex[4] = 'N';

	baseIndex['A'+1] = 3;	baseIndex[6] = 'T';  
	baseIndex['T'+1] = 0;	baseIndex[7] = 'G';
	baseIndex['C'+1] = 2;	baseIndex[8] = 'C';
	baseIndex['G'+1] = 1;	baseIndex[9] = 'A';
	baseIndex['N'+1] = 4;	baseIndex[10] = 'N';

}

void seqOut(FILE *fp, char *cnsSeq, int numCns, int index, char *seqName)
{
	int i;
	fprintf(fp, "@%s_%08d\n%s\n", seqName, index, cnsSeq);
	fprintf(fp, "+%s_%08d\n", seqName, index);
	fprintf(fp, "!&&&&&&&&&&&&&&&&&&&");  
	for(i=0;i<numCns-40;++i) putc(40+041,fp);
	fprintf(fp, "&&&&&&&&&&&&&&&&&&&&\n");  
}

void seqExtR(KmerDB *kmerP, SeqDB *seqP, ReadInf *readInfP, ReptKmerId *reptKmerIdP, ExtCntlD *extCntlP)
{
	int i, j;
	int kmerL = kmerP->kmerL;
	char *cnsSeqP;
	int numCns;
	int uniqKmerI;
	int initDir;

	int nContig = extCntlP->nContig;
	cnsSeqP = extCntlP->cnsSeq;

	int nRept = reptKmerIdP->nRept;
	for(i=0;i<nRept;++i){

		uniqKmerI = reptKmerIdP->uKmerId[i].next;
		initDir = -1;
		extCntlP->nContig=nContig;
		extCntlP->reptWord=reptKmerIdP->uKmerId[i].word;
		numCns = kmerWalkingE(cnsSeqP, uniqKmerI, kmerP, seqP, readInfP, reptKmerIdP, extCntlP, -1);

		if(numCns<=2 && extCntlP->breakKmer==reptKmerIdP->uKmerId[i].word){
			extCntlP->reptWord = -1;
			initDir = -initDir;
			extCntlP->nContig=nContig;
			numCns = kmerWalkingE(cnsSeqP, uniqKmerI, kmerP, seqP, readInfP, reptKmerIdP, extCntlP, initDir);
		}
		if(numCns != 0){
			reverseComplement(cnsSeqP, (long) numCns);
		}

		initDir = -initDir;
		extCntlP->nContig=nContig;
		extCntlP->reptWord=-1;
		int numCns2;
		numCns2 = kmerWalkingE(cnsSeqP+numCns+1, uniqKmerI, kmerP, seqP, readInfP, reptKmerIdP, extCntlP, initDir);
		char *cnsSeqTmp;
		char *cnsSeqTmp2;
		if(numCns>kmerL){
			cnsSeqTmp = cnsSeqP+numCns;
			cnsSeqTmp2 = cnsSeqP+numCns+kmerL+1;
			if(numCns2-kmerL>0){
				for(j=0;j<numCns2-kmerL;++j) *cnsSeqTmp++=*cnsSeqTmp2++;
				*cnsSeqTmp = '\0';
				numCns += numCns2 - kmerL;
			}
		}
		else {
			if(numCns2==0) continue;
			cnsSeqTmp = cnsSeqP;
			cnsSeqTmp2 = cnsSeqP+numCns+1;
			for(j=0;j<numCns2;++j) *cnsSeqTmp++=*cnsSeqTmp2++;
			*cnsSeqTmp = '\0';
			numCns = numCns2;
		}
		if(numCns > 50) seqOut(extCntlP->fpOut, extCntlP->cnsSeq, numCns, i, "extra_seq");
		++nContig;
	}
}

void seqExtM(KmerDB *kmerP, SeqDB *seqP, ReadInf *readInfP, ReptKmerId *reptKmerIdP,ExtCntlD *extCntlP)
{
	int i, n;
	int numCns, readId;
	int uniqKmerI;
	int ri=0; float rf;

	if(extCntlP->nExtRead==0)extCntlP->nExtRead = kmerP->totUsedKmer*25/3000;
	if(extCntlP->nExtRead >= seqP->nRead) extCntlP->useAllRead = 1;

	int *readSel = (int *) malloc(sizeof(int)*extCntlP->nExtRead);  
	char *usedKmer = (char *) malloc(sizeof(char)*kmerP->nUKmer);

	if(extCntlP->useAllRead == 0){
		srand(0);
		memset(readSel,-1,extCntlP->nExtRead);
		memset(usedKmer,0,kmerP->nUKmer);
	        for(i=0;i<extCntlP->nExtRead;++i){
			n=0;
	                while(1){ 
				if(++n>10*extCntlP->nExtRead){
					extCntlP->useAllRead = 1;
					break;
				}
	                        rf = 1.0*rand()/RAND_MAX; 
	                        ri=(int) (rf*(seqP->nRead-1));  
				if(readInfP[ri].kmerI == -1) continue;
				if(usedKmer[readInfP[ri].kmerI] == 0) break;
	                }
			if(extCntlP->useAllRead == 1) break;
			usedKmer[readInfP[ri].kmerI] = 1;
	                readSel[i] = ri;
	        }
	}
	if(extCntlP->useAllRead == 1){
		memset(readSel,-1,extCntlP->nExtRead);
		memset(usedKmer,0,kmerP->nUKmer);
		n = 0;
		for(i=0;i<seqP->nRead;++i){
			if(readInfP[i].kmerI == -1) continue;
			if(usedKmer[readInfP[i].kmerI] != 0) continue;
			readSel[n++] = i;
			usedKmer[readInfP[i].kmerI] = 1;
		}
		extCntlP->nExtRead = n;
		fprintf(stderr,"Warning: the number of output sequence required has been changed into %d!\n",extCntlP->nExtRead);
	}
	free(usedKmer);

	reptKmerIdP->nRept = 0;
	reptKmerIdP->uKmerId = (ReptKmer *) malloc(sizeof(ReptKmer)*extCntlP->nExtRead*2);

	int nContig = extCntlP->nContig;

	qsort(readSel, extCntlP->nExtRead, sizeof(int), intCmp);

	for(i=0;i<extCntlP->nExtRead;i++){
		readId = readSel[i];
		uniqKmerI = readInfP[readId].kmerI;

		extCntlP->nContig=nContig;
		numCns = kmerWalking(extCntlP->cnsSeq, uniqKmerI, kmerP, seqP, readInfP, reptKmerIdP, extCntlP, 1);
#ifdef RELEASE
		if(numCns==0 && numCns==1) continue;
#endif
		if(numCns > 50) seqOut(extCntlP->fpOut, extCntlP->cnsSeq, numCns, i, "extend_seq");
		++nContig;
	}
}

